'
' xlink.spin
'
' An implementation of the XMOS serial external link protocol.
'
' Uses two COGS and 4 pins
'
' Transfers at a rate of 4 PASM instructions per bit.
' That's 5Mb/s on a standard 80MHz Propeller setup.
'
'
' Design criteria:
'
' 1) As fast as possible. Need I say.
'
' 3) Minimal use of pins. A valuable resource on a Propeller. The XMOS
'    parallel protocol requires 10 pins.
'
' 4) Please forgive the use of 2 COGs. This may seem wastefull for so little code
'    but a single COG solution would substantially slower. Given that the intent is to export
'    functionality (ethernet, USB, SD card etc) to an external chip this is a
'    resonable trade off.
'
' Copile with BST (or HomeSpun I think) as the @@@ operator is used.
'
' Michael Rychlik
'
' v0.1   02-12-2009  First attempt.
'
'
CON
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 6_553_600

  baud          = 115_200


  buffer_size = 64
  buffer_mask = buffer_size - 1

OBJ
  term : "FullDuplexSerial_rr004"



VAR
  long  cog_rx                     'cog flag/id
  long  cog_tx                     'cog flag/id

DAT
rx_head  long  0
rx_tail  long  0
tx_head  long  0
tx_tail  long  0

VAR
  long  buffer_ptr

  word  rx_buffer[buffer_size]
  word  tx_buffer[buffer_size]



PUB main
  'Start the terminal
  term.start(31,30,0,baud)

  'Sign on message
  term.str(string("Xlink for the Propeller"))
  crlf

  'Delay so you can get PropTerminal active.
  waitcnt(clkfreq*1 + cnt)

  received_token := $11111111

  start

  'Display the received XLINK token
  repeat
    term.hex(received_token, 8)
    crlf
    waitcnt(clkfreq*1 + cnt)



PUB start : ok
  stop

  'Start up the XLNK receiver cog
  term.str(string("Starting receiver"))
  crlf
  cog_rx := cognew(@enter_rx, 0) + 1
  ok := cog_rx + 1

  'Delay
  waitcnt(clkfreq*1 + cnt)

  'Start up the XLNK transmitter cog
  term.str(string("Starting transmitter"))
  crlf
  cog_tx := cognew(@enter_tx, 0)
  ok := cog_tx + 1



PUB stop
  if cog_rx
    cogstop(cog_rx~)
  if cog_tx
    cogstop(cog_tx~)


PUB tx(tx_token)
' Send a token (may wait for room in buffer)
' Bits 0-7 of token are data, bit 8 is conrol token flag
  repeat until (tx_tail <> (tx_head + 1) & buffer_mask)
  tx_buffer[tx_head] := tx_token
  tx_head := (tx_head + 1) & buffer_mask


PRI crlf
  term.tx(13)
  term.tx(10)


DAT
'XLINK Receiver
                        org     0
enter_rx
token_rx                mov     rtoken, #0

                        waitpne rest_state, pins_rx     'Wait for wires to leave the rest state
                        mov     in_0, ina               'Get wires from P0 and P1
                        test    in_0, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_10000000    'Wire 1 set so set data bit 7

                        nop                             'Delay to get subsequent samples nearer the
                        nop                             'middle of bit periods.

                        mov     in_1, ina               'Get wires from P0 and P1
                        xor     in_0, in_1              'Get wire changes
                        test    in_0, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_01000000    'Wire 1 change so set data bit 6

                        mov     in_0, ina               'Get wires from P0 and P1
                        xor     in_1, in_0              'Get wire changes
                        test    in_1, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00100000    'Wire 1 change so set data bit 5

                        mov     in_1, ina               'Get wires from P0 and P1
                        xor     in_0, in_1              'Get wire changes
                        test    in_0, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00010000    'Wire 1 change so set data bit 4

                        mov     in_0, ina               'Get wires from P0 and P1
                        xor     in_1, in_0              'Get wire changes
                        test    in_1, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00001000    'Wire 1 change so set data bit 3

                        mov     in_1, ina               'Get wires from P0 and P1
                        xor     in_0, in_1              'Get wire changes
                        test    in_0, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00000100    'Wire 1 change so set data bit 2

                        mov     in_0, ina               'Get wires from P0 and P1
                        xor     in_1, in_0              'Get wire changes
                        test    in_1, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00000010    'Wire 1 change so set data bit 1

                        mov     in_1, ina               'Get wires from P0 and P1
                        xor     in_0, in_1              'Get wire changes
                        test    in_0, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%0_00000001    'Wire 1 change so set data bit 0

                        mov     in_0, ina               'Get wires from P0 and P1
                        xor     in_1, in_0              'Get wire changes
                        test    in_1, #%1000 wz         'Test for wire 1 change
        if_nz           or      rtoken, #%1_00000000    'Wire 1 change so set "control" bit

                        mov     in_1, ina
                        test    in_1, pins_rx wz        'Check wires for rest state (0,0) at end  of token
        if_nz           mov     rtoken, minus_one       'Indicate parity error if need be.

                        'Do something with the token here
                        wrlong  rtoken, received_token_addr
                        jmp     #token_rx

in_0                    long    0
in_1                    long    0
rtoken                  long    0
rest_state              long    0
pins_rx                 long    %00000000_00000000_00000000_00001100
minus_one               long    $FFFFFFFF
received_token_addr     long    @@@received_token
junk                    long    $666

                        fit     $1F0

received_token          long    $F0F0F0F0


DAT
'XLINK Transmitter
                        org     0

enter_tx                mov     token_count, one_million
                        mov     outa, #0
                        mov     dira, #%1100            'Set P3 and P3 to output for wires 1 and 0 resp.

tx_loop                 sub     token_count, #1
                        cmp     token_count, #0 wz
        if_z            mov     token_count, one_million
        if_z            add     ttoken, #1

                        call    #token_tx

                        nop
                        nop
                        nop
                        nop
                        jmp     #tx_loop




token_tx                test    ttoken, #%0_10000000 wz 'Get data bit 7 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0.
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_01000000 wz 'Get data bit 6 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00100000 wz 'Get data bit 5 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00010000 wz 'Get data bit 4 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00001000 wz 'Get data bit 3 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00000100 wz 'Get data bit 2 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00000010 wz 'Get data bit 1 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%0_00000001 wz 'Get data bit 0 into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        test    ttoken, #%1_00000000 wz 'Get "control" indicator into carry
        if_nz           mov     out_1, wire_1           'To toggle wire 1...
        if_z            mov     out_1, wire_0           '...or wire 0
                        xor     outa, out_1             'Toggle selected wire

                        nop                             'Bit time delay before comming to rest.
                        nop
                        nop
                        mov     outa, #%00              'Set wires to rest (0,0) state.
token_tx_ret            ret


ttoken                  long    0
out_1                   long    0
pins_tx                 long    %00000000_00000000_00000000_00001100 'The TX output pins
wire_0                  long    %00000000_00000000_00000000_00000100 'Toggle this pin to TX a 0
wire_1                  long    %00000000_00000000_00000000_00001000 'Toggle this pin to TX a 1
tx_head_p               long    @@@tx_head
tx_tail_p               long    @@@tx_tail
'tx_buffer_p             long    @@@tx_buffer
temp_1                  long    0
temp_2                  long    0
temp_3                  long    0

one_million             long    1_000_000
token_count             long    0
                        fit     $1F0

